@@ -1,9 +1,17 @@
 use std::{fmt, num::NonZeroU8};
 
 pub enum ParseConfigError {
-    NotFound,
+    NoSuchConfig,
     InvalidValue,
 }
+impl fmt::Display for ParseConfigError {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            Self::NoSuchConfig => f.write_str("no such config"),
+            Self::InvalidValue => f.write_str("invalid config value"),
+        }
+    }
+}
 
 macro_rules! config_values {
     ($($name:ident: $type:ty = $default:expr,)*) => {
@@ -20,7 +28,7 @@ macro_rules! config_values {
                         Ok(value) => self.$name = value,
                         Err(_) => return Err(ParseConfigError::InvalidValue),
                     },)*
-                    _ => return Err(ParseConfigError::NotFound),
+                    _ => return Err(ParseConfigError::NoSuchConfig),
                 }
                 Ok(())
             }
@@ -61,10 +69,10 @@ config_values! {
     tab_size: NonZeroU8 = NonZeroU8::new(4).unwrap(),
     indent_with_tabs: bool = true,
 
-    visual_empty: u8 = b'~',
-    visual_space: u8 = b'.',
-    visual_tab_first: u8 = b'|',
-    visual_tab_repeat: u8 = b' ',
+    visual_empty: char = '~',
+    visual_space: char = '.',
+    visual_tab_first: char = '|',
+    visual_tab_repeat: char = ' ',
 
     completion_min_len: u8 = 3,
     picker_max_height: u8 = 8,
