@@ -3,7 +3,9 @@ use std::{error::Error, fmt, str::Chars};
 use crate::{
     buffer::BufferHandle,
     buffer_position::BufferRange,
+    buffer_view::BufferViewHandle,
     client::ClientHandle,
+    cursor::Cursor,
     platform::Key,
     serialization::{DeserializeError, Deserializer, Serialize, Serializer},
 };
@@ -15,13 +17,24 @@ pub struct EditorEventText {
 }
 impl EditorEventText {
     pub fn as_str<'a>(&self, events: &'a EditorEventQueue) -> &'a str {
-        &events.read.texts[self.from as usize..self.to as _]
+        &events.read.texts[self.from as usize..self.to as usize]
+    }
+}
+
+#[derive(Clone, Copy)]
+pub struct EditorEventCursors {
+    from: u32,
+    to: u32,
+}
+impl EditorEventCursors {
+    pub fn as_cursors<'a>(&self, events: &'a EditorEventQueue) -> &'a [Cursor] {
+        &events.read.cursors[self.from as usize..self.to as usize]
     }
 }
 
 pub enum EditorEvent {
     Idle,
-    BufferLoad {
+    BufferRead {
         handle: BufferHandle,
     },
     BufferInsertText {
@@ -33,15 +46,16 @@ pub enum EditorEvent {
         handle: BufferHandle,
         range: BufferRange,
     },
-    BufferSave {
+    BufferWrite {
         handle: BufferHandle,
         new_path: bool,
     },
     BufferClose {
         handle: BufferHandle,
     },
-    ClientChangeBufferView {
-        handle: ClientHandle,
+    FixCursors {
+        handle: BufferViewHandle,
+        cursors: EditorEventCursors,
     },
 }
 
@@ -49,6 +63,7 @@ pub enum EditorEvent {
 struct EventQueue {
     events: Vec<EditorEvent>,
     texts: String,
+    cursors: Vec<Cursor>,
 }
 
 #[derive(Default)]
@@ -57,17 +72,22 @@ pub struct EditorEventQueue {
     write: EventQueue,
 }
 impl EditorEventQueue {
-    pub fn flip(&mut self) {
+    pub(crate) fn flip(&mut self) {
         self.read.events.clear();
         self.read.texts.clear();
         std::mem::swap(&mut self.read, &mut self.write);
     }
 
-    pub fn enqueue(&mut self, event: EditorEvent) {
+    pub(crate) fn enqueue(&mut self, event: EditorEvent) {
         self.write.events.push(event);
     }
 
-    pub fn enqueue_buffer_insert(&mut self, handle: BufferHandle, range: BufferRange, text: &str) {
+    pub(crate) fn enqueue_buffer_insert(
+        &mut self,
+        handle: BufferHandle,
+        range: BufferRange,
+        text: &str,
+    ) {
         let from = self.write.texts.len();
         self.write.texts.push_str(text);
         let text = EditorEventText {
@@ -80,6 +100,18 @@ impl EditorEventQueue {
             text,
         });
     }
+
+    pub(crate) fn enqueue_fix_cursors(&mut self, handle: BufferViewHandle, cursors: &[Cursor]) {
+        let from = self.write.cursors.len();
+        self.write.cursors.extend_from_slice(cursors);
+        let cursors = EditorEventCursors {
+            from: from as _,
+            to: self.write.cursors.len() as _,
+        };
+        self.write
+            .events
+            .push(EditorEvent::FixCursors { handle, cursors });
+    }
 }
 
 pub struct EditorEventIter(usize);
@@ -89,13 +121,9 @@ impl EditorEventIter {
     }
 
     pub fn next<'a>(&mut self, queue: &'a EditorEventQueue) -> Option<&'a EditorEvent> {
-        if self.0 < queue.read.events.len() {
-            let event = &queue.read.events[self.0];
-            self.0 += 1;
-            Some(event)
-        } else {
-            None
-        }
+        let event = queue.read.events.get(self.0)?;
+        self.0 += 1;
+        Some(event)
     }
 }
 
@@ -121,26 +149,25 @@ pub struct KeyParseAllError {
 }
 impl fmt::Display for KeyParseAllError {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        f.write_fmt(format_args!("{} at index: {}", self.error, self.index))
+        write!(f, "{} at char: {}", self.error, self.index)
     }
 }
 impl Error for KeyParseAllError {}
 
 pub struct KeyParser<'a> {
-    raw: &'a str,
     chars: Chars<'a>,
+    raw: &'a str,
 }
 impl<'a> KeyParser<'a> {
     pub fn new(raw: &'a str) -> Self {
         Self {
-            raw,
             chars: raw.chars(),
+            raw,
         }
     }
 }
 impl<'a> Iterator for KeyParser<'a> {
     type Item = Result<Key, KeyParseAllError>;
-
     fn next(&mut self) -> Option<Self::Item> {
         if self.chars.as_str().is_empty() {
             return None;
@@ -154,15 +181,14 @@ impl<'a> Iterator for KeyParser<'a> {
                     .next_back()
                     .map(|(i, _)| i)
                     .unwrap_or(0);
-
+                self.raw = "";
                 Some(Err(KeyParseAllError { index, error }))
             }
         }
     }
 }
 
-fn parse_key(chars: &mut impl Iterator<Item = char>) -> Result<Key, KeyParseError> {
-    #[inline]
+fn parse_key(chars: &mut Chars) -> Result<Key, KeyParseError> {
     fn next(chars: &mut impl Iterator<Item = char>) -> Result<char, KeyParseError> {
         match chars.next() {
             Some(c) => Ok(c),
@@ -170,7 +196,6 @@ fn parse_key(chars: &mut impl Iterator<Item = char>) -> Result<Key, KeyParseErro
         }
     }
 
-    #[inline]
     fn consume(chars: &mut impl Iterator<Item = char>, c: char) -> Result<(), KeyParseError> {
         let next = next(chars)?;
         if c == next {
@@ -180,7 +205,6 @@ fn parse_key(chars: &mut impl Iterator<Item = char>) -> Result<Key, KeyParseErro
         }
     }
 
-    #[inline]
     fn consume_str(chars: &mut impl Iterator<Item = char>, s: &str) -> Result<(), KeyParseError> {
         for c in s.chars() {
             consume(chars, c)?
@@ -188,92 +212,92 @@ fn parse_key(chars: &mut impl Iterator<Item = char>) -> Result<Key, KeyParseErro
         Ok(())
     }
 
-    let key = match next(chars)? {
+    match next(chars)? {
         '<' => match next(chars)? {
             'b' => {
                 consume_str(chars, "ackspace>")?;
-                Key::Backspace
+                Ok(Key::Backspace)
             }
             's' => {
                 consume_str(chars, "pace>")?;
-                Key::Char(' ')
+                Ok(Key::Char(' '))
             }
             'e' => match next(chars)? {
                 'n' => match next(chars)? {
                     't' => {
                         consume_str(chars, "er>")?;
-                        Key::Enter
+                        Ok(Key::Enter)
                     }
                     'd' => {
                         consume(chars, '>')?;
-                        Key::End
+                        Ok(Key::End)
                     }
-                    c => return Err(KeyParseError::InvalidCharacter(c)),
+                    c => Err(KeyParseError::InvalidCharacter(c)),
                 },
                 's' => {
                     consume_str(chars, "c>")?;
-                    Key::Esc
+                    Ok(Key::Esc)
                 }
-                c => return Err(KeyParseError::InvalidCharacter(c)),
+                c => Err(KeyParseError::InvalidCharacter(c)),
             },
             'l' => {
                 consume(chars, 'e')?;
                 match next(chars)? {
                     's' => {
                         consume_str(chars, "s>")?;
-                        Key::Char('<')
+                        Ok(Key::Char('<'))
                     }
                     'f' => {
                         consume_str(chars, "t>")?;
-                        Key::Left
+                        Ok(Key::Left)
                     }
-                    c => return Err(KeyParseError::InvalidCharacter(c)),
+                    c => Err(KeyParseError::InvalidCharacter(c)),
                 }
             }
             'g' => {
                 consume_str(chars, "reater>")?;
-                Key::Char('>')
+                Ok(Key::Char('>'))
             }
             'r' => {
                 consume_str(chars, "ight>")?;
-                Key::Right
+                Ok(Key::Right)
             }
             'u' => {
                 consume_str(chars, "p>")?;
-                Key::Up
+                Ok(Key::Up)
             }
             'd' => match next(chars)? {
                 'o' => {
                     consume_str(chars, "wn>")?;
-                    Key::Down
+                    Ok(Key::Down)
                 }
                 'e' => {
                     consume_str(chars, "lete>")?;
-                    Key::Delete
+                    Ok(Key::Delete)
                 }
-                c => return Err(KeyParseError::InvalidCharacter(c)),
+                c => Err(KeyParseError::InvalidCharacter(c)),
             },
             'h' => {
                 consume_str(chars, "ome>")?;
-                Key::Home
+                Ok(Key::Home)
             }
             'p' => {
                 consume_str(chars, "age")?;
                 match next(chars)? {
                     'u' => {
                         consume_str(chars, "p>")?;
-                        Key::PageUp
+                        Ok(Key::PageUp)
                     }
                     'd' => {
                         consume_str(chars, "own>")?;
-                        Key::PageDown
+                        Ok(Key::PageDown)
                     }
-                    c => return Err(KeyParseError::InvalidCharacter(c)),
+                    c => Err(KeyParseError::InvalidCharacter(c)),
                 }
             }
             't' => {
                 consume_str(chars, "ab>")?;
-                Key::Tab
+                Ok(Key::Tab)
             }
             'f' => {
                 let c = next(chars)?;
@@ -284,49 +308,42 @@ fn parse_key(chars: &mut impl Iterator<Item = char>) -> Result<Key, KeyParseErro
                             Some(d1) => {
                                 consume(chars, '>')?;
                                 let n = d0 * 10 + d1;
-                                Key::F(n as _)
-                            }
-                            None => {
-                                if c == '>' {
-                                    Key::F(d0 as _)
-                                } else {
-                                    return Err(KeyParseError::InvalidCharacter(c));
-                                }
+                                Ok(Key::F(n as _))
                             }
+                            None => match c {
+                                '>' => Ok(Key::F(d0 as _)),
+                                _ => Err(KeyParseError::InvalidCharacter(c)),
+                            },
                         }
                     }
-                    None => return Err(KeyParseError::InvalidCharacter(c)),
+                    None => Err(KeyParseError::InvalidCharacter(c)),
                 }
             }
             'c' => {
                 consume(chars, '-')?;
                 let c = next(chars)?;
-                let key = if c.is_ascii_alphanumeric() {
-                    Key::Ctrl(c)
+                if c.is_ascii_alphanumeric() {
+                    consume(chars, '>')?;
+                    Ok(Key::Ctrl(c))
                 } else {
-                    return Err(KeyParseError::InvalidCharacter(c));
-                };
-                consume(chars, '>')?;
-                key
+                    Err(KeyParseError::InvalidCharacter(c))
+                }
             }
             'a' => {
                 consume(chars, '-')?;
                 let c = next(chars)?;
-                let key = if c.is_ascii_alphanumeric() {
-                    Key::Alt(c)
+                if c.is_ascii_alphanumeric() {
+                    consume(chars, '>')?;
+                    Ok(Key::Alt(c))
                 } else {
-                    return Err(KeyParseError::InvalidCharacter(c));
-                };
-                consume(chars, '>')?;
-                key
+                    Err(KeyParseError::InvalidCharacter(c))
+                }
             }
-            c => return Err(KeyParseError::InvalidCharacter(c)),
+            c => Err(KeyParseError::InvalidCharacter(c)),
         },
-        c @ '>' => return Err(KeyParseError::InvalidCharacter(c)),
-        c => Key::Char(c),
-    };
-
-    Ok(key)
+        '>' => Err(KeyParseError::InvalidCharacter('>')),
+        c => Ok(Key::Char(c)),
+    }
 }
 
 impl fmt::Display for Key {
@@ -345,13 +362,13 @@ impl fmt::Display for Key {
             Key::PageDown => f.write_str("<pagedown>"),
             Key::Tab => f.write_str("<tab>"),
             Key::Delete => f.write_str("<delete>"),
-            Key::F(n) => f.write_fmt(format_args!("<f{}>", n)),
+            Key::F(n) => write!(f, "<f{}>", n),
             Key::Char(' ') => f.write_str("<space>"),
             Key::Char('<') => f.write_str("<less>"),
             Key::Char('>') => f.write_str("<greater>"),
-            Key::Char(c) => f.write_fmt(format_args!("{}", c)),
-            Key::Ctrl(c) => f.write_fmt(format_args!("<c-{}>", c)),
-            Key::Alt(c) => f.write_fmt(format_args!("<a-{}>", c)),
+            Key::Char(c) => write!(f, "{}", c),
+            Key::Ctrl(c) => write!(f, "<c-{}>", c),
+            Key::Alt(c) => write!(f, "<a-{}>", c),
             Key::Esc => f.write_str("<esc>"),
         }
     }
@@ -438,19 +455,22 @@ where
 pub enum ServerEvent<'a> {
     Display(&'a [u8]),
     Suspend,
-    CommandOutput(&'a str),
-    Request(&'a str),
+    StdoutOutput(&'a [u8]),
 }
 impl<'a> ServerEvent<'a> {
-    pub const fn header_len() -> usize {
+    pub const fn bytes_variant_header_len() -> usize {
         1 + std::mem::size_of::<u32>()
     }
 
-    pub fn serialize_display_header(buf: &mut [u8]) {
-        buf[0] = 0;
-        let len = buf.len() as u32 - Self::header_len() as u32;
+    pub fn serialize_bytes_variant_header(&self, buf: &mut [u8]) {
+        buf[0] = match self {
+            Self::Display(_) => 0,
+            Self::Suspend => unreachable!(),
+            Self::StdoutOutput(_) => 2,
+        };
+        let len = buf.len() as u32 - Self::bytes_variant_header_len() as u32;
         let len_buf = len.to_le_bytes();
-        buf[1..Self::header_len()].copy_from_slice(&len_buf);
+        buf[1..Self::bytes_variant_header_len()].copy_from_slice(&len_buf);
     }
 }
 impl<'de> Serialize<'de> for ServerEvent<'de> {
@@ -464,13 +484,9 @@ impl<'de> Serialize<'de> for ServerEvent<'de> {
                 display.serialize(serializer);
             }
             Self::Suspend => 1u8.serialize(serializer),
-            Self::CommandOutput(output) => {
+            Self::StdoutOutput(bytes) => {
                 2u8.serialize(serializer);
-                output.serialize(serializer);
-            }
-            Self::Request(request) => {
-                3u8.serialize(serializer);
-                request.serialize(serializer);
+                bytes.serialize(serializer);
             }
         }
     }
@@ -487,22 +503,48 @@ impl<'de> Serialize<'de> for ServerEvent<'de> {
             }
             1 => Ok(Self::Suspend),
             2 => {
-                let output = Serialize::deserialize(deserializer)?;
-                Ok(Self::CommandOutput(output))
-            }
-            3 => {
-                let request = Serialize::deserialize(deserializer)?;
-                Ok(Self::Request(request))
+                let bytes = Serialize::deserialize(deserializer)?;
+                Ok(Self::StdoutOutput(bytes))
             }
             _ => Err(DeserializeError::InvalidData),
         }
     }
 }
 
+#[derive(Clone, Copy)]
+pub enum TargetClient {
+    Sender,
+    Focused,
+}
+impl<'de> Serialize<'de> for TargetClient {
+    fn serialize<S>(&self, serializer: &mut S)
+    where
+        S: Serializer,
+    {
+        match self {
+            Self::Sender => 0u8.serialize(serializer),
+            Self::Focused => 1u8.serialize(serializer),
+        }
+    }
+
+    fn deserialize<D>(deserializer: &mut D) -> Result<Self, DeserializeError>
+    where
+        D: Deserializer<'de>,
+    {
+        let discriminant = u8::deserialize(deserializer)?;
+        match discriminant {
+            0 => Ok(Self::Sender),
+            1 => Ok(Self::Focused),
+            _ => Err(DeserializeError::InvalidData),
+        }
+    }
+}
+
 pub enum ClientEvent<'a> {
-    Command(ClientHandle, &'a str),
-    Key(ClientHandle, Key),
-    Resize(ClientHandle, u16, u16),
+    Key(TargetClient, Key),
+    Resize(u16, u16),
+    Command(TargetClient, &'a str),
+    StdinInput(TargetClient, &'a [u8]),
 }
 impl<'de> Serialize<'de> for ClientEvent<'de> {
     fn serialize<S>(&self, serializer: &mut S)
@@ -510,22 +552,26 @@ impl<'de> Serialize<'de> for ClientEvent<'de> {
         S: Serializer,
     {
         match self {
-            Self::Command(client_handle, command) => {
+            Self::Key(target, key) => {
                 0u8.serialize(serializer);
-                client_handle.serialize(serializer);
-                command.serialize(serializer);
-            }
-            Self::Key(client_handle, key) => {
-                1u8.serialize(serializer);
-                client_handle.serialize(serializer);
+                target.serialize(serializer);
                 serialize_key(*key, serializer);
             }
-            Self::Resize(client_handle, width, height) => {
-                2u8.serialize(serializer);
-                client_handle.serialize(serializer);
+            Self::Resize(width, height) => {
+                1u8.serialize(serializer);
                 width.serialize(serializer);
                 height.serialize(serializer);
             }
+            Self::Command(target, command) => {
+                2u8.serialize(serializer);
+                target.serialize(serializer);
+                command.serialize(serializer);
+            }
+            Self::StdinInput(target, bytes) => {
+                3u8.serialize(serializer);
+                target.serialize(serializer);
+                bytes.serialize(serializer);
+            }
         }
     }
 
@@ -536,20 +582,24 @@ impl<'de> Serialize<'de> for ClientEvent<'de> {
         let discriminant = u8::deserialize(deserializer)?;
         match discriminant {
             0 => {
-                let handle = Serialize::deserialize(deserializer)?;
-                let command = Serialize::deserialize(deserializer)?;
-                Ok(Self::Command(handle, command))
+                let target = Serialize::deserialize(deserializer)?;
+                let key = deserialize_key(deserializer)?;
+                Ok(Self::Key(target, key))
             }
             1 => {
-                let handle = Serialize::deserialize(deserializer)?;
-                let key = deserialize_key(deserializer)?;
-                Ok(Self::Key(handle, key))
-            }
-            2 => {
-                let handle = Serialize::deserialize(deserializer)?;
                 let width = Serialize::deserialize(deserializer)?;
                 let height = Serialize::deserialize(deserializer)?;
-                Ok(Self::Resize(handle, width, height))
+                Ok(Self::Resize(width, height))
+            }
+            2 => {
+                let target = Serialize::deserialize(deserializer)?;
+                let command = Serialize::deserialize(deserializer)?;
+                Ok(Self::Command(target, command))
+            }
+            3 => {
+                let target = Serialize::deserialize(deserializer)?;
+                let bytes = Serialize::deserialize(deserializer)?;
+                Ok(Self::StdinInput(target, bytes))
             }
             _ => Err(DeserializeError::InvalidData),
         }
@@ -577,8 +627,14 @@ impl ClientEventIter {
         }
     }
 
-    pub fn finish(&self, receiver: &mut ClientEventReceiver) {
+    pub fn finish(self, receiver: &mut ClientEventReceiver) {
         receiver.bufs[self.buf_index].drain(..self.read_len);
+        std::mem::forget(self);
+    }
+}
+impl Drop for ClientEventIter {
+    fn drop(&mut self) {
+        panic!("forgot to call 'finish' on ClientEventIter");
     }
 }
 
@@ -591,7 +647,7 @@ impl ClientEventReceiver {
     pub fn receive_events(&mut self, client_handle: ClientHandle, bytes: &[u8]) -> ClientEventIter {
         let buf_index = client_handle.into_index();
         if buf_index >= self.bufs.len() {
-            self.bufs.resize_with(buf_index + 1, Default::default);
+            self.bufs.resize_with(buf_index + 1, Vec::new);
         }
         let buf = &mut self.bufs[buf_index];
         buf.extend_from_slice(bytes);
@@ -724,7 +780,7 @@ mod tests {
         }
 
         let client_handle = ClientHandle::from_index(0).unwrap();
-        let event = ClientEvent::Key(client_handle, Key::Char(CHAR));
+        let event = ClientEvent::Key(TargetClient::Sender, Key::Char(CHAR));
         let mut bytes = Vec::new();
         for _ in 0..EVENT_COUNT {
             event.serialize(&mut bytes);
@@ -751,4 +807,3 @@ mod tests {
         assert_eq!(EVENT_COUNT, event_count);
     }
 }
-
